
/**************************************************************************

Copyright 1996 Richard Mairesse

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appears in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation. No responsibility is assumed for the use
or reliability of this software.

Author:
  Richard Mairesse                                Tel: 0032 81 725009
  Facultes Universitaires Notre-Dame de la Paix   Fax: 0032 81 725023
  Centre de Calcul                                Richard.Mairesse@fundp.ac.be
  rue Grandgagnage 21
  5000 Namur
  Belgium

**************************************************************************/

#include <stdio.h>
#include <math.h>
#include <X11/StringDefs.h>
#include <Xm/XmP.h>
#include <X11/CoreP.h>
#include "JoystickP.h"


static void left_move(),right_move(),down_move(),up_move(),
            big_left_move(),big_right_move(),big_down_move(),big_up_move(),
            middle_move(),select_joystick();
static void Initialize();
static void Redisplay();
static void Resize();
static void Destroy();
static Boolean SetValues();


static char defaultTranslations[] = 
       "<Btn1Motion>: select()          \n\
           <Key>Left: left_move()       \n\
          <Key>Right: right_move()      \n\
           <Key>Down: down_move()       \n\
             <Key>Up: up_move()         \n\
             c<Key>b: left_move()       \n\
             c<Key>f: right_move()      \n\
             c<Key>p: down_move()       \n\
             c<Key>n: up_move()         \n\
             c<Key>d: middle_move()     \n\
             s<Key>b: big_left_move()   \n\
             s<Key>f: big_right_move()  \n\
             s<Key>p: big_down_move()   \n\
             s<Key>n: big_up_move()     \n\
             s<Key>d: middle_move()     \n\
          <Btn1Down>: select()";

static XtActionsRec actionsList[] = {
  {"select",     (XtActionProc) select_joystick},
  {"left_move",  (XtActionProc) left_move},
  {"right_move", (XtActionProc) right_move},
  {"down_move",  (XtActionProc) down_move},
  {"up_move",    (XtActionProc) up_move},
  {"big_left_move",  (XtActionProc) big_left_move},
  {"big_right_move", (XtActionProc) big_right_move},
  {"big_down_move",  (XtActionProc) big_down_move},
  {"big_up_move",    (XtActionProc) big_up_move},
  {"middle_move",(XtActionProc) middle_move},
};


static XtResource resources[] = {
  {XtNselect,XtCCallback,XtRCallback,sizeof(caddr_t),
   XtOffset(XrmJoystickWidget,joystick.select),XtRCallback,NULL},
  {XtNxMinimum,XtCMin,XtRInt,sizeof(int),
   XtOffset(XrmJoystickWidget,joystick.x_min),XtRString,"-100"},
  {XtNxMaximum,XtCMax,XtRInt,sizeof(int),
   XtOffset(XrmJoystickWidget,joystick.x_max),XtRString,"100"},
  {XtNyMinimum,XtCMin,XtRInt,sizeof(int),
   XtOffset(XrmJoystickWidget,joystick.y_min),XtRString,"-100"},
  {XtNyMaximum,XtCMax,XtRInt,sizeof(int),
   XtOffset(XrmJoystickWidget,joystick.y_max),XtRString,"100"},
  {XtNxValue,XtCValue,XtRInt,sizeof(int),
   XtOffset(XrmJoystickWidget,joystick.x_value),XtRString,"0"},
  {XtNyValue,XtCValue,XtRInt,sizeof(int),
   XtOffset(XrmJoystickWidget,joystick.y_value),XtRString,"0"},
  {XtNxProcessingDirection,XtCProcessingDirection,XtRBoolean,sizeof(Boolean),
   XtOffset(XrmJoystickWidget,joystick.x_proc),XtRImmediate,(XtPointer)False},
  {XtNyProcessingDirection,XtCProcessingDirection,XtRBoolean,sizeof(Boolean),
   XtOffset(XrmJoystickWidget,joystick.y_proc),XtRImmediate,(XtPointer)False},
  {XtNstickPercentage,XtCStickPercentage,XtRFloat,sizeof(float),
   XtOffset(XrmJoystickWidget,joystick.stick_percentage),XtRString,"30"},
  {XtNshadowEffect,XtCShadowEffect,XtRBoolean,sizeof(Boolean),
   XtOffset(XrmJoystickWidget,joystick.shadow_effect),XtRImmediate,
   (XtPointer)True}
};


XrmJoystickClassRec XrmjoystickClassRec = {
  {
/*    (WidgetClass) &widgetClassRec,*/
    (WidgetClass) &xmPrimitiveClassRec,
    "Joystick",
    sizeof(XrmJoystickRec),
    NULL,
    NULL,
    FALSE,
    Initialize,
    NULL,
    XtInheritRealize,
    actionsList,
    XtNumber(actionsList),
    resources,
    XtNumber(resources),
    NULLQUARK,
    TRUE,
    TRUE,
    TRUE,
    TRUE,
    Destroy,
    Resize,
    Redisplay,
    SetValues,
    NULL,
    XtInheritSetValuesAlmost,
    NULL,
    NULL,
    XtVersion,
    NULL,
    defaultTranslations,
    NULL,
    NULL,
    NULL,
  },
  {
    _XtInherit,
    _XtInherit,
    XtInheritTranslations,
    NULL,
    NULL,
    0,
    NULL,
  },
  {
    0,
  }
};


WidgetClass XrmjoystickWidgetClass = (WidgetClass) &XrmjoystickClassRec;


static void coord_widget_stick(w)
     XrmJoystickWidget w;
{w->joystick.x_stick = 
       (int)(w->joystick.a * (float)w->joystick.x_value + w->joystick.tx);
 w->joystick.y_stick = 
       (int)(w->joystick.d * (float)w->joystick.y_value + w->joystick.ty);}


static void coord_stick_widget(w)
     XrmJoystickWidget w;
{w->joystick.x_value = 
       (int)((float)w->joystick.x_stick / w->joystick.a + w->joystick.tpx);
 w->joystick.y_value = 
       (int)((float)w->joystick.y_stick / w->joystick.d + w->joystick.tpy);}


static unsigned int get_depth(w)
Widget w;
{
Window r;
int x,y;
unsigned int wd,ht,bw,depth;
XGetGeometry(XtDisplay(w),RootWindowOfScreen(XtScreen(w)),
             &r,&x,&y,&wd,&ht,&bw,&depth);
return depth;
}


static void Initialize(request,new)
     XrmJoystickWidget request,new;
{
XGCValues values;
XtGCMask valueMask;
if (request->core.width == 0)
  new->core.width = 100;
if (request->core.height == 0)
  new->core.height = 100;
if (new->joystick.x_min >= new->joystick.x_max) {
  XtWarning("X Maximum must be greater than the X Minimum");
  new->joystick.x_min = -100;
  new->joystick.x_max = 100;}
if (new->joystick.y_min >= new->joystick.y_max) {
  XtWarning("Y Maximum must be greater than the Y Minimum");
  new->joystick.y_min = -100;
  new->joystick.y_max = 100;}
if (new->joystick.x_value > new->joystick.x_max) {
  XtWarning("X supplied value must be less than the X Maximum");
  new->joystick.x_value = new->joystick.x_max;}
if (new->joystick.x_value < new->joystick.x_min) {
  XtWarning("X supplied value must be greater than the X Minimum");
  new->joystick.x_value = new->joystick.x_min;}
if (new->joystick.y_value > new->joystick.y_max) {
  XtWarning("Y supplied value must be less than the Y Maximum");
  new->joystick.y_value = new->joystick.y_max;}
if (new->joystick.stick_percentage < 5) {
  XtWarning("Stick percentage must be greater than 5 \%");
  new->joystick.stick_percentage = 5;}
if (new->joystick.stick_percentage > 100) {
  XtWarning("Stick percentage must be less than 100 \%");
  new->joystick.stick_percentage = 100;}
valueMask = GCForeground | GCBackground | GCClipXOrigin | GCClipYOrigin;
values.foreground = new->primitive.foreground;
values.background = new->core.background_pixel;
values.clip_x_origin = 0;values.clip_y_origin = 0;
new->joystick.gc = XCreateGC(XtDisplay(new),RootWindowOfScreen(XtScreen(new)),
                             valueMask,&values);
valueMask = GCForeground | GCBackground;
new->joystick.gcoffscreen = XCreateGC(XtDisplay(new),
                   RootWindowOfScreen(XtScreen(new)),valueMask,&values);
new->joystick.cmask = XCreatePixmap(XtDisplay(new),
    RootWindowOfScreen(XtScreen(new)),new->core.width,new->core.height,
    (unsigned int)1);
new->joystick.gcmask = XCreateGC(XtDisplay(new),
                                  new->joystick.cmask,NULL,NULL);
new->joystick.offscreen = XCreatePixmap(XtDisplay(new),
    RootWindowOfScreen(XtScreen(new)),
    new->core.width,new->core.height,get_depth(new));
Resize(new);
}


static void Destroy(w)
     XrmJoystickWidget w;
{
XFreePixmap(XtDisplay(w),w->joystick.offscreen);
XFreePixmap(XtDisplay(w),w->joystick.cmask);
XtDestroyGC(w->joystick.gc);
XtDestroyGC(w->joystick.gcoffscreen);
XtDestroyGC(w->joystick.gcmask);
XtRemoveAllCallbacks(w,XtNselect,w->joystick.select);
}


static void Resize(w)
     XrmJoystickWidget w;
{
XGCValues values;
XtGCMask valueMask;
if (w->core.width < w->core.height) 
 {w->joystick.radius_stick = 
      (int)(w->joystick.stick_percentage * (float)w->core.width/200.0);
  w->joystick.radius_base  = w->core.width /2 - w->joystick.radius_stick;}
else
 {w->joystick.radius_stick = 
      (int)(w->joystick.stick_percentage * (float)w->core.height/200.0);
  w->joystick.radius_base  = w->core.height /2 - w->joystick.radius_stick;}
w->joystick.drs = 2* w->joystick.radius_stick;
w->joystick.cx =    w->core.width     / 2;
w->joystick.cy =    w->core.height    / 2;
if (w->joystick.x_proc)
   {w->joystick.x_inf = w->joystick.cx + w->joystick.radius_base;
    w->joystick.x_sup = w->joystick.cx - w->joystick.radius_base;}
else
   {w->joystick.x_inf = w->joystick.cx - w->joystick.radius_base;
    w->joystick.x_sup = w->joystick.cx + w->joystick.radius_base;}
if (w->joystick.y_proc)
   {w->joystick.y_inf = w->joystick.cy - w->joystick.radius_base;
    w->joystick.y_sup = w->joystick.cy + w->joystick.radius_base;}
else
  {w->joystick.y_inf = w->joystick.cy + w->joystick.radius_base;
   w->joystick.y_sup = w->joystick.cy - w->joystick.radius_base;}
w->joystick.a =    (float)(w->joystick.x_sup - w->joystick.x_inf)/
                   (float)(w->joystick.x_max - w->joystick.x_min);
w->joystick.d =    (float)(w->joystick.y_sup - w->joystick.y_inf)/
                   (float)(w->joystick.y_max - w->joystick.y_min);
w->joystick.tx =   (float)(w->joystick.x_max * w->joystick.x_inf - 
                    w->joystick.x_sup * w->joystick.x_min)/
                   (float)(w->joystick.x_max - w->joystick.x_min);
w->joystick.ty =   (float)(w->joystick.y_max * w->joystick.y_inf - 
                    w->joystick.y_sup * w->joystick.y_min)/
                   (float)(w->joystick.y_max - w->joystick.y_min);
w->joystick.tpx =  (float)(w->joystick.x_sup * w->joystick.x_min - 
                    w->joystick.x_max * w->joystick.x_inf)/
                   (float)(w->joystick.x_sup - w->joystick.x_inf);
w->joystick.tpy =  (float)(w->joystick.y_sup * w->joystick.y_min - 
                    w->joystick.y_max * w->joystick.y_inf)/
                   (float)(w->joystick.y_sup - w->joystick.y_inf);
coord_widget_stick(w);
XFreePixmap(XtDisplay(w),w->joystick.cmask);
w->joystick.cmask = XCreatePixmap(XtDisplay(w),
    RootWindowOfScreen(XtScreen(w)),w->core.width,w->core.height,
    (unsigned int)1);
XtDestroyGC(w->joystick.gcmask);
w->joystick.gcmask = XCreateGC(XtDisplay(w),w->joystick.cmask,NULL,NULL);
XFreePixmap(XtDisplay(w),w->joystick.offscreen);
w->joystick.offscreen = XCreatePixmap(XtDisplay(w),
    RootWindowOfScreen(XtScreen(w)),
    w->joystick.drs,w->joystick.drs,get_depth(w));
}


static void change_circular_mask(w,i)
     XrmJoystickWidget w;
     int i;
{
unsigned int r;
int cx,cy;
if (i == 0) 
  {r = w->joystick.radius_base;
   cx = w->joystick.cx;
   cy = w->joystick.cy;}
else
  {r = w->joystick.radius_stick;
   cx = w->joystick.x_stick;
   cy = w->joystick.y_stick;}
XSetForeground(XtDisplay(w),w->joystick.gcmask,(unsigned int)0);
XSetBackground(XtDisplay(w),w->joystick.gcmask,(unsigned int)1);
XFillRectangle(XtDisplay(w),w->joystick.cmask,w->joystick.gcmask,
               0,0,w->core.width,w->core.height);
XSetForeground(XtDisplay(w),w->joystick.gcmask,(unsigned int)1);
XSetBackground(XtDisplay(w),w->joystick.gcmask,(unsigned int)0);
XFillArc(XtDisplay(w),w->joystick.cmask,w->joystick.gcmask,cx - r,cy - r,
         (unsigned int)(2*r),(unsigned int)(2*r),0,360*64);
XSetClipMask(XtDisplay(w),w->joystick.gc,w->joystick.cmask);
}



/*
static void draw_base(w)
     XrmJoystickWidget w;
{
XSetClipMask(XtDisplay(w),w->joystick.gc,None);
XSetForeground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XSetBackground(XtDisplay(w),w->joystick.gc,w->primitive.foreground);
XFillRectangle(XtDisplay(w),XtWindow(w),w->joystick.gc,
               0,0,w->core.width,w->core.height);
}
*/


static void draw_base(w)
     XrmJoystickWidget w;
{
int r,rp,dr,dum;
XPoint points[8];
XSetClipMask(XtDisplay(w),w->joystick.gc,None);
if (w->core.width < w->core.height) 
     dum = (w->core.width /2  - w->joystick.radius_base)/5;
else dum = (w->core.height /2 - w->joystick.radius_base)/5;
r = w->joystick.radius_base + dum;
rp = r + dum;
dr = 2*r;
points[0].x = points[3].x = w->joystick.cx - rp;
points[0].y = points[3].y = w->joystick.cy - rp;
points[1].x = points[6].x = w->joystick.cx + rp;
points[1].y = points[6].y = w->joystick.cy - rp;
points[2].x = points[5].x = w->joystick.cx - rp;
points[2].y = points[5].y = w->joystick.cy + rp;
points[4].x = points[7].x = w->joystick.cx + rp;
points[4].y = points[7].y = w->joystick.cy + rp;
XSetForeground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XSetBackground(XtDisplay(w),w->joystick.gc,w->primitive.foreground);
XFillRectangle(XtDisplay(w),XtWindow(w),w->joystick.gc,
             0,0,(unsigned int)w->core.width,(unsigned int)w->core.height);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.top_shadow_color);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XFillPolygon(XtDisplay(w),XtWindow(w),w->joystick.gc,points + 1,3,
             Convex,CoordModeOrigin);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.bottom_shadow_color);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XFillPolygon(XtDisplay(w),XtWindow(w),w->joystick.gc,points + 4,3,
             Convex,CoordModeOrigin);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.foreground);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XDrawLines(XtDisplay(w),XtWindow(w),w->joystick.gc,points,8,CoordModeOrigin);
XSetForeground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XSetBackground(XtDisplay(w),w->joystick.gc,w->primitive.foreground);
XFillArc(XtDisplay(w),XtWindow(w),w->joystick.gc,
             w->joystick.cx - r,w->joystick.cy - r,dr,dr,0,360*64);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.foreground);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XDrawArc(XtDisplay(w),XtWindow(w),w->joystick.gc,
             w->joystick.cx - r,w->joystick.cy - r,dr,dr,0,360*64);
}


void draw_cavity(w)
     XrmJoystickWidget w;
{
int r,dr,tr,lw;
XGCValues getgc;
double f;
r = w->joystick.radius_base;
dr = 2 * r;
tr = 3 * r;
f = (double)tr*sqrt(2.0)/2.0;
change_circular_mask(w,0);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.bottom_shadow_color);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XFillRectangle(XtDisplay(w),XtWindow(w),w->joystick.gc,0,0,
               w->core.width,w->core.height);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.top_shadow_color);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XFillArc(XtDisplay(w),XtWindow(w),w->joystick.gc,
         w->joystick.cx - tr + (int)f,w->joystick.cy - tr + (int)f,
         (unsigned int)(2*tr),(unsigned int)(2*tr),0,360*64);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.foreground);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XGetGCValues(XtDisplay(w),w->joystick.gc,(unsigned long)GCLineWidth,&getgc);
lw = getgc.line_width;
if (lw == 0) lw = 1;
XDrawArc(XtDisplay(w),XtWindow(w),w->joystick.gc,
w->joystick.cx - r + lw,w->joystick.cy - r + lw,dr - 2*lw,dr - 2*lw,0,360*64);
}


void draw_stick(w)
     XrmJoystickWidget w;
{
int r,dr,tr,lw;
XGCValues getgc;
double f;
r = w->joystick.radius_stick;
dr = 2 * r;
tr = 3 * r;
f = (double)tr*sqrt(2.0)/2.0;
XCopyArea(XtDisplay(w),w->joystick.offscreen,XtWindow(w),
          w->joystick.gcoffscreen,0,0,(unsigned int)dr,(unsigned int)dr,
          w->joystick.x_offscreen,w->joystick.y_offscreen);
XCopyArea(XtDisplay(w),XtWindow(w),w->joystick.offscreen,
          w->joystick.gcoffscreen,
          w->joystick.x_offscreen = w->joystick.x_stick - r,
	  w->joystick.y_offscreen = w->joystick.y_stick - r,
          (unsigned int)dr,(unsigned int)dr,0,0);
change_circular_mask(w,1);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.bottom_shadow_color);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XFillRectangle(XtDisplay(w),XtWindow(w),w->joystick.gc,0,0,
               w->core.width,w->core.height);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.top_shadow_color);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XFillArc(XtDisplay(w),XtWindow(w),w->joystick.gc,
         w->joystick.x_stick - tr - (int)f,w->joystick.y_stick - tr - (int)f,
         (unsigned int)(2*tr),(unsigned int)(2*tr),0,360*64);
XSetForeground(XtDisplay(w),w->joystick.gc,w->primitive.foreground);
XSetBackground(XtDisplay(w),w->joystick.gc,w->core.background_pixel);
XGetGCValues(XtDisplay(w),w->joystick.gc,(unsigned long)GCLineWidth,&getgc);
lw = getgc.line_width;
if (lw == 0) lw = 1;
XDrawArc(XtDisplay(w),XtWindow(w),w->joystick.gc,
         w->joystick.x_stick - r + lw,w->joystick.y_stick - r + lw,
         dr - 2 * lw,dr - 2 * lw,0,360*64);
}


static void Redisplay(w,event,region)
     XrmJoystickWidget w;
     XEvent *event;
     Region region;
{
if (w->core.visible)
  {draw_base(w);
   draw_cavity(w);
   XCopyArea(XtDisplay(w),XtWindow(w),w->joystick.offscreen,
             w->joystick.gcoffscreen,
             w->joystick.x_offscreen = w->joystick.x_stick - 
                                       w->joystick.radius_stick,
	     w->joystick.y_offscreen = w->joystick.y_stick - 
                                       w->joystick.radius_stick,
             (unsigned int)w->joystick.drs,(unsigned int)w->joystick.drs,0,0);
   draw_stick(w);}
}


static Boolean SetValues(current,request,new)
     XrmJoystickWidget current,request,new;
{
Boolean redraw   = FALSE;
Boolean resize_w = FALSE;
if (new->joystick.x_min >= new->joystick.x_max) {
  XtWarning("X Maximum must be greater than the X Minimum");
  new->joystick.x_min = -100;
  new->joystick.x_max = 100;}
if (new->joystick.y_min >= new->joystick.y_max) {
  XtWarning("Y Maximum must be greater than the Y Minimum");
  new->joystick.y_min = -100;
  new->joystick.y_max = 100;}
if (new->joystick.x_value > new->joystick.x_max) {
  XtWarning("X supplied value must be less than the X Maximum");
  new->joystick.x_value = new->joystick.x_max;}
if (new->joystick.x_value < new->joystick.x_min) {
  XtWarning("X supplied value must be greater than the X Minimum");
  new->joystick.x_value = new->joystick.x_min;}
if (new->joystick.y_value > new->joystick.y_max) {
  XtWarning("Y supplied value must be less than the Y Maximum");
  new->joystick.y_value = new->joystick.y_max;}
if (new->joystick.stick_percentage < 5) {
  XtWarning("Stick percentage must be greater than 5 \%");
  new->joystick.stick_percentage = 5;}
if (new->joystick.stick_percentage > 100) {
  XtWarning("Stick percentage must be less than 100 \%");
  new->joystick.stick_percentage = 100;}
if ((new->joystick.x_min != current->joystick.x_min) ||
    (new->joystick.x_max != current->joystick.x_max) ||
    (new->joystick.y_min != current->joystick.y_min) ||
    (new->joystick.y_max != current->joystick.y_max))   
  {resize_w = True;
   redraw = TRUE;}
if (new->joystick.stick_percentage != current->joystick.stick_percentage)
  {resize_w = True;
   redraw = TRUE;}
if (new->joystick.x_proc != current->joystick.x_proc)
  {resize_w = True;
   redraw = TRUE;}
if (new->joystick.y_proc != current->joystick.y_proc)
  {resize_w = True;
   redraw = TRUE;}
if ((new->joystick.x_value != current->joystick.x_value) ||
    (new->joystick.x_value != current->joystick.x_value))
  {coord_widget_stick(new);
   resize_w = True;
   redraw = True;}
if (resize_w) Resize(new);
return(redraw);
}


static void CallAllCallbacks(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
XrmJoystickCallbackStruct JCS;
coord_stick_widget(w);
draw_stick(w);
JCS.reason = 0;
JCS.event = event;
JCS.x_value = w->joystick.x_value;
JCS.y_value = w->joystick.y_value;
XtCallCallbacks(w,XtNselect,&JCS);
}


static void CallAllCallbacks_bis(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
XrmJoystickCallbackStruct JCS;
coord_widget_stick(w);
draw_stick(w);
JCS.reason = 0;
JCS.event = event;
JCS.x_value = w->joystick.x_value;
JCS.y_value = w->joystick.y_value;
XtCallCallbacks(w,XtNselect,&JCS);
}



/* Circular boundaries version with ButtonPress attraction */
/*
static void select_joystick(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
int x,y,r,d;
double c1,c2;
if (event->type == ButtonPress)
  {x = event->xbutton.x;y = event->xbutton.y;}
if (event->type == MotionNotify)
  {x = event->xmotion.x;y = event->xmotion.y;}
c1 = (double)w->joystick.cx - (double)x;
c2 = (double)w->joystick.cy - (double)y;
d = (int)(sqrt(c1*c1 + c2*c2));
r = w->joystick.radius_base;
if (d <= r)
  {w->joystick.x_stick = x;
   w->joystick.y_stick = y;}
else
  {w->joystick.x_stick = w->joystick.cx + (x - w->joystick.cx)*r/d;
   w->joystick.y_stick = w->joystick.cy + (y - w->joystick.cy)*r/d;}
CallAllCallbacks(w,event);
}
*/

static void check_xy(w,x,y,s,t)
     XrmJoystickWidget w;
     int x,y,*s,*t;
{
float xp,yp,r;
r = (float)w->joystick.radius_base;
xp = (float)(x - w->joystick.cx);
yp = (float)(y - w->joystick.cy);

if (xp*xp >= yp*yp)
  {if (xp >= r) {*s = (int)r + w->joystick.cx; 
                 *t = (int)(r*yp/xp) + w->joystick.cy; return;}
   if (xp <= -r) {*s = w->joystick.cx - (int)r; 
                  *t = w->joystick.cy - (int)(r*yp/xp); return;}}
else
  {if (yp >= r) {*s = (int)(r*xp/yp) + w->joystick.cx;
                 *t = (int)r + w->joystick.cy; return;}
   if (yp <= -r) {*s = w->joystick.cx - (int)(r*xp/yp);
                  *t = w->joystick.cy - (int)r; return;}}
*s = x; *t = y;
}


/* Square boundaries version with no attractive ButtonPress */
static void select_joystick(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
int x,y,d,dx,dy,t1,t2;
double c1,c2;
if (event->type == ButtonPress)
  {x = event->xbutton.x;
   y = event->xbutton.y;
   c1 = (double)w->joystick.x_stick - (double)x;
   c2 = (double)w->joystick.y_stick - (double)y;
   d =  (int)(sqrt(c1*c1 + c2*c2));
   if (d <= w->joystick.radius_stick)
     {w->joystick.x_tmp = x - w->joystick.x_stick;
      w->joystick.y_tmp = y - w->joystick.y_stick;
      return;}
   else
     {check_xy(w,x,y,&t1,&t2);
      w->joystick.x_stick = t1;
      w->joystick.y_stick = t2;
      w->joystick.x_tmp = 0;
      w->joystick.y_tmp = 0;
      CallAllCallbacks(w,event);
      return;}}
if (event->type == MotionNotify)
  {x = event->xmotion.x;y = event->xmotion.y;
   dx = w->joystick.x_tmp;
   dy = w->joystick.y_tmp;
   check_xy(w,x - dx,y - dy,&t1,&t2);
   w->joystick.x_stick = t1;
   w->joystick.y_stick = t2;
   CallAllCallbacks(w,event);}
}


static void left_move(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
if (w->joystick.x_value > w->joystick.x_min) --w->joystick.x_value;
CallAllCallbacks_bis(w,event);
}


static void big_left_move(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
w->joystick.x_value = w->joystick.x_min;
CallAllCallbacks_bis(w,event);
}


static void right_move(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
if (w->joystick.x_value < w->joystick.x_max) ++w->joystick.x_value;
CallAllCallbacks_bis(w,event);
}


static void big_right_move(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
w->joystick.x_value = w->joystick.x_max;
CallAllCallbacks_bis(w,event);
}


static void down_move(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
if (w->joystick.y_value > w->joystick.y_min) --w->joystick.y_value;
CallAllCallbacks_bis(w,event);
}


static void big_down_move(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
w->joystick.y_value = w->joystick.y_min;
CallAllCallbacks_bis(w,event);
}


static void up_move(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
if (w->joystick.y_value < w->joystick.y_max) ++w->joystick.y_value;
CallAllCallbacks_bis(w,event);
}


static void big_up_move(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
w->joystick.y_value = w->joystick.y_max;
CallAllCallbacks_bis(w,event);
}


static void middle_move(w,event)
     XrmJoystickWidget w;
     XEvent *event;
{
w->joystick.x_value = (w->joystick.x_min + w->joystick.x_max)/2;
w->joystick.y_value = (w->joystick.y_min + w->joystick.y_max)/2;
CallAllCallbacks_bis(w,event);
}









